import click
import os
import sys
import re
import yaml
import operator
import functools
import subprocess
import collections.abc
import numpy as np
import tqdm
from collections import defaultdict
from ._version import *


def CheckAllIsFile(paths):
    r'''
Check that all paths point to existing files.
Uses: ``os.path.isfile``.

:param list paths: List of file paths.
:throw: IOError
    '''

    for path in paths:
        if not os.path.isfile(path):
            raise IOError('"{0:s}" does not exist'.format(path))


def DirNames(paths, return_unique=True):
    r'''
Get the ``os.path.dirname`` of all file paths.

:param list paths: List of file paths.
:param bool return_unique: Filter duplicates.
:return: List of dirnames.
    '''

    ret = [os.path.dirname(path) for path in paths]

    if not return_unique:
        return ret

    return list(set(ret))


def OverWrite(paths, force=False):
    r'''
List files that will be overwritten.

:param list paths: List of file paths.
:param bool force: If ``True`` the user is prompted to overwrite.
:return: List of overwritten files.
    '''

    ret = [path for path in paths if os.path.isfile(path)]

    if not force:
        return ret

    print('Files exist:')
    print('\n'.join(ret))
    if not click.confirm('Overwrite?'):
        raise IOError('Cancelled')

    return ret


def _FlattenList_detail(data):
    r'''
See https://stackoverflow.com/a/17485785/2646505
    '''

    for item in data:
        if isinstance(item, collections.abc.Iterable) and not isinstance(item, str):
            for x in FlattenList(item):
                yield x
        else:
            yield item


def FlattenList(data):
    r'''
Flatten a nested list to a one dimensional list.
    '''
    return list(_FlattenList_detail(data))


def _Squash_detail(data, parent_key='', sep='_'):
    r'''
See https://stackoverflow.com/a/6027615/2646505
    '''

    items = []

    for k, v in data.items():

        new_key = parent_key + sep + k if parent_key else k

        if isinstance(v, collections.abc.MutableMapping):
            items.extend(_Squash_detail(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))

    return dict(items)


def Squash(data):
    r'''
Squash a dictionary to a single list.
    '''

    return FlattenList(list(_Squash_detail(data).values()))


def _ExecCommand(cmd, verbose=False):
    r'''
Run command, optionally verbose command and its output, and return output.

:type cmd: str
:param cmd: The command to run.

:type verbose: bool
:param verbose: Print command and its output.
    '''

    if verbose:
        print(cmd)

    ret = subprocess.check_output(cmd, shell=True).decode('utf-8')

    if verbose:
        print(ret)

    return ret


def YamlRead(filename):
    r'''
Read YAML file and return its content as ``list`` or ``dict``.
    '''

    if not os.path.isfile(filename):
        raise IOError('"{0:s} does not exist'.format(filename))

    with open(filename, 'r') as file:
        return yaml.load(file.read(), Loader=yaml.FullLoader)


def YamlDump(filename, data, force=False):
    r'''
Dump data to YAML file.

:type filename: str
:param filename: The output filename.

:type data: list, dict
:param data: The data to dump.

:type force: bool, optional
:param force: Do not prompt to overwrite file.
    '''

    dirname = os.path.dirname(filename)

    if not force:
        if os.path.isfile(filename):
            if not click.confirm('Overwrite "{0:s}"?'.format(filename)):
                raise IOError('Cancelled')
        elif not os.path.isdir(dirname) and len(dirname) > 0:
            if not click.confirm('Create "{0:s}"?'.format(os.path.dirname(filename))):
                raise IOError('Cancelled')

    if not os.path.isdir(dirname) and len(dirname) > 0:
        os.makedirs(os.path.dirname(filename))

    with open(filename, 'w') as file:
        ret = yaml.dump(data, file)


def YamlPrint(data):
    r'''
Print data formatted as YAML.
    '''
    print(yaml.dump(data, default_flow_style=False, default_style=''))


def YamlGetItem(filename, key=[]):
    r'''
Get an item from a YAML file.
Optionally the key to the item can be specified as a list. E.g.
*   ``[]`` for a YAML file containing only a list.
*   ``['foo']`` for a plain YAML file.
*   ``['key', 'to', foo']`` for a YAML file with nested items.
    '''

    data = YamlRead(filename)

    if len(key) == 0 and type(data) != list:
        raise IOError('Specify key for "{0:s}"'.format(filename))

    if len(key) > 0:
        try:
            return functools.reduce(operator.getitem, key, data)
        except:
            raise IOError('"{0:s}" not in "{1:s}"'.format('/'.join(key), filename))

    return data


def PrefixPaths(prefix, files):
    r'''
Add prefix to a list of filenames.
Skip if all paths are absolute paths.

:param str prefix: The prefix.
:param list files: List of paths.
:return: List of paths.
    '''

    isabs = [os.path.isabs(file) for file in files]

    if any(isabs) and not all(isabs):
        raise IOError('Specify either relative or absolute files-paths')

    if all(isabs):
        return files

    return [os.path.normpath(os.path.join(prefix, file)) for file in files]


def ChangeRootOfRelativePaths(files, old_root, new_root, in_place=False):
    r'''
Change the root of relative paths.
Skip if all paths are absolute paths.

If ``in_place = True`` the input list is modified 'in place' (and a pointer to it is returned),
otherwise a new list is returned.
    '''

    isabs = [os.path.isabs(file) for file in files]

    if any(isabs) and not all(isabs):
        raise IOError('Specify either relative or absolute files-paths')

    if all(isabs):
        return files

    if not in_place:
        return [os.path.relpath(os.path.abspath(os.path.join(old_root, file)), new_root) for file in files]

    for i in range(len(files)):
        files[i] = os.path.relpath(os.path.abspath(os.path.join(old_root, files[i])), new_root)

    return files


def IsOnRemote(host, source, verbose=False):
    r'''
Check if a file exists on a remote system. Uses ``ssh``.
    '''

    cmd = 'ssh {host:s} test -f "{source:s}" && echo found || echo not found'.format(
        host=host, source=source)

    ret = _ExecCommand(cmd, args['--verbose'])

    if ret == 'found':
        return True

    return False


def CopyFromRemote(host, source, dest, verbose=False):
    r'''
Copy a file from a remote system. Uses ``scp -p``.
    '''

    cmd = 'scp -p {host:s}:{source:s} {dest:s}'.format(
        host=host, source=source, dest=dest)

    _ExecCommand(cmd, verbose)


def CopyToRemote(host, source, dest, verbose=False):
    r'''
Copy a file to a remote system. Uses ``scp -p``.
    '''

    cmd = 'scp -p {source:s} {host:s}:{dest:s}'.format(
        host=host, source=source, dest=dest)

    _ExecCommand(cmd, verbose)


def RsyncFromRemote(
    hostname,
    source_dir,
    dest_dir,
    tempfilename,
    files,
    force=False,
    verbose=False,
    progress=True):
    r'''
Copy files to a remote system using ``rsync -a --files-from``.
    '''

    assert type(tempfilename) == str

    if not force:
        if os.path.isfile(tempfilename):
            if not click.confirm('Overwrite "{0:s}"?'.format(tempfilename)):
                raise IOError('Cancelled')

    open(tempfilename, 'w').write('\n'.join(files))

    # Run without printing output

    if not progress:

        cmd = 'rsync -a --files-from="{files:s}" {hostname:s}:{source_dir:s} {dest_dir:s}'.format(
            hostname=hostname, source_dir=source_dir, dest_dir=dest_dir, files=tempfilename)

        return _ExecCommand(cmd, verbose)

    # Run while printing output

    cmd = 'rsync -aP --files-from="{files:s}" {hostname:s}:{source_dir:s} {dest_dir:s}'.format(
        hostname=hostname, source_dir=source_dir, dest_dir=dest_dir, files=tempfilename)

    if verbose:
        print(cmd)

    pbar = tqdm.tqdm(total=len(files))
    sbar = tqdm.tqdm(unit='B', unit_scale=True)

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)

    for line in iter(process.stdout.readline, b''):
        line = line.decode("utf-8")
        if re.match(r'(.*)(xf)([e]?)(r\#)([0-9])(.*)(to\-ch)([e]?[c]?)(k\=)([0-9])(.*)', line):
            e = int(list(filter(None, line.split(" ")))[-6].replace(",", ""))
            pbar.update()
            sbar.update(e)


def RsyncToRemote(
    hostname,
    source_dir,
    dest_dir,
    tempfilename,
    files,
    force=False,
    verbose=False,
    progress=True):
    r'''
Copy files from a remote system using ``rsync -a --files-from``.
    '''

    assert type(tempfilename) == str

    if not force:
        if os.path.isfile(tempfilename):
            if not click.confirm('Overwrite "{0:s}"?'.format(tempfilename)):
                raise IOError('Cancelled')

    open(tempfilename, 'w').write('\n'.join(files))

    # Run without printing output

    if not progress:

        cmd = 'rsync -a --files-from="{files:s}" {source_dir:s} {hostname:s}:{dest_dir:s}'.format(
            hostname=hostname, source_dir=source_dir, dest_dir=dest_dir, files=tempfilename)

        return _ExecCommand(cmd, verbose)

    # Run while printing output

    cmd = 'rsync -aP --files-from="{files:s}" {source_dir:s} {hostname:s}:{dest_dir:s}'.format(
        hostname=hostname, source_dir=source_dir, dest_dir=dest_dir, files=tempfilename)

    if verbose:
        print(cmd)

    pbar = tqdm.tqdm(total=len(files))
    sbar = tqdm.tqdm(unit='B', unit_scale=True)

    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)

    for line in iter(process.stdout.readline, ''):
        line = line.decode("utf-8")
        if re.match(r'(.*)(xfr\#)([0-9])(.*)(to\-chk\=)([0-9])(.*)', line):
            e = int(list(filter(None, line.split(" ")))[-6].replace(",", ""))
            pbar.update()
            sbar.update(e)


def MakeDir(dirname, force=False):
    r'''
Create a directory if it does not yet exist.

:param bool force: Create directories without prompt.
    '''

    if len(dirname) == 0:
        return 0

    if os.path.isdir(dirname):
        return 0

    if not force:
        print('mkdir -p {0:s}'.format(dirname))
        if not click.confirm('Proceed?'):
            raise IOError('Cancelled')

    os.makedirs(dirname)


def _to_tree(d):
    t = defaultdict(list)
    for a, *b in d:
        t[a].append(b)
    return {a: None if not (k := list(filter(None, b))) else _to_tree(k) for a, b in t.items()}

def _get_deepest_paths(d, c=[]):
    r'''
See https://stackoverflow.com/a/66211932/2646505
    '''
    for a, b in d.items():
        if b is None:
            yield '/'.join(c + [a])
        else:
            yield from _get_deepest_paths(b, c + [a])

def GetDeepestPaths(paths):
    r'''
Return list with only the deepest paths.

For example::

    GetDeepestPaths(["foo/bar/dir", "foo/bar"])
    >>> ["foo/bar/dir"]

:type paths: list of str
:param paths: List of paths.

:rtype: list of str
:return: List of paths.
    '''

    return list(_get_deepest_paths(_to_tree([i.split('/') for i in paths])))


def MakeDirs(dirnames, force=False):
    r'''
Same as :py:mod:`shelephant.MakeDir` but for list of directories.

:param bool force: Create directories without prompt.
    '''

    dirnames = [dirname for dirname in dirnames if not os.path.isdir(dirname)]

    if len(dirnames) == 0:
        return 0

    dirnames = list(set(dirnames))
    dirnames = sorted(GetDeepestPaths(dirnames))

    if not force:
        for dirname in dirnames:
            print('mkdir -p {0:s}'.format(dirname))
        if not click.confirm('Proceed?'):
            raise IOError('Cancelled')

    for dirname in dirnames:
        os.makedirs(dirname)


def GetSHA256(filename, size = 2 ** 10):
    r'''
Get SHA256 for a file.
    '''

    import hashlib

    h = hashlib.sha256()

    with open(filename, 'rb') as f:
        for byte_block in iter(lambda: f.read(size * h.block_size), b""):
            h.update(byte_block)
        return h.hexdigest()


def GetChecksums(filepaths, yaml_hostinfo=None, hybrid=False, progress=False):
    r'''
Compute the checksums for ``filepaths``.

:param list filepaths: List of file-paths.

:param set yaml_hostinfo:
    File-path of a host-info file (see :py:mod:`shelephant.cli.hostinfo`).
    If specified the checksums are **not** computed, but exclusively read from the
    host-file. The user is responsible for keeping them up-to-date.

:param bool hybrid:
    If ``True``, the function first tries to read from ``yaml_hostinfo``, and then
    computes missing items on the fly.

:param bool progress: Show a progress-bar.

:return:
    List of checksums, of same length as ``filepaths``.
    The entry is ``None`` if no checksum was found/read.
    '''

    n = len(filepaths)
    ret = [None for i in range(n)]

    # Compute

    if not yaml_hostinfo:

        for i in tqdm.trange(n, disable=not progress, desc='Processing'):
            if os.path.isfile(filepaths[i]):
                ret[i] = GetSHA256(filepaths[i])

        return ret

    # Read pre-computed

    data = YamlRead(yaml_hostinfo)
    files = data['files']
    prefix = data['prefix']
    check_sums = data['checksum']
    check_paths = PrefixPaths(prefix, files)

    sorter = np.argsort(filepaths)
    source_paths = np.array(filepaths)[sorter]

    i = np.argsort(check_paths)
    check_paths = np.array(check_paths)[i]
    check_sums = np.array(check_sums)[i]

    test = np.in1d(source_paths, check_paths)
    idx = np.searchsorted(check_paths, source_paths)
    idx = np.where(test, idx, 0)
    ret = np.where(test, check_sums[idx], None)
    out = np.empty_like(ret)
    out[sorter] = ret
    ret = list(out)

    if hybrid:
        for i in tqdm.trange(n, disable=not progress, desc='Processing'):
            if ret[i] is None:
                if os.path.isfile(filepaths[i]):
                    ret[i] = GetSHA256(filepaths[i])

    return ret


def ShelephantCopy(
    copy_function,
    files,
    src_dir,
    dest_dir,
    checksum = False,
    quiet = False,
    force = False,
    print_details = True,
    print_summary = True,
    print_all = False,
    theme_name = 'none',
    yaml_hostinfo_src = None,
    yaml_hostinfo_dest = None):
    r'''
Copy/move files.

:param function copy_function: Function to perform the copy. E.g. ``os.rename``.

:type files: list of str
:param files: Filenames (will be prepended by ``src_dir`` and ``dest_dir``).

:param str src_dir: The source directory.

:param str dest_dir: The destination directory.

:param bool checksum: Use checksum to skip files that are the same.

:param bool quiet: Proceed without printing progress.

:param bool force: Continue without prompt.

:param bool print_details: Print copy details.

:param bool print_summary: Print copy summary.

:param bool print_all: If ``False`` auto-truncation of long output is applied.

:type theme_name: str or None
:param theme_name: The name of the color-theme. See :py:mod:`shelephant.Theme`.

:type yaml_hostinfo_src: str, optional
:param yaml_hostinfo_src:
    Filename of hostinfo for the source, see :py:mod:`shelephant.cli.hostinfo`.
    Specify these files *only* to use precomputed checksums.

:type yaml_hostinfo_dest: str, optional
:param yaml_hostinfo_dest:
    Filename of hostinfo for the destination, see :py:mod:`shelephant.cli.hostinfo`.
    Specify these files *only* to use precomputed checksums.
    '''

    assert type(files) == list

    if len(files) == 0:
        if not quiet:
            print('Nothing to do')
        return

    src = PrefixPaths(src_dir, files)
    dest = PrefixPaths(dest_dir, files)
    n = len(src)
    overwrite = [False for i in range(n)]
    create = [False for i in range(n)]
    skip = [False for i in range(n)]
    theme = Theme(theme_name.lower())

    for file in src:
        if not os.path.isfile(file):
            raise IOError('Input file "{0:s}" does not exists'.format(file))

    if not os.path.isdir(dest_dir) and len(dest_dir) > 0:

        create = [True for i in range(n)]

    else:

        if checksum == True:
            src_checksums = GetChecksums(src, yaml_hostinfo_src, progress=not quiet)
            dest_checksums = GetChecksums(dest, yaml_hostinfo_dest, progress=not quiet)

        for i in range(n):
            if os.path.isfile(dest[i]):
                if checksum:
                    if (src_checksums[i] == dest_checksums[i]) and (src_checksums[i] is not None):
                        skip[i] = True
                        continue
                overwrite[i] = True
                continue
            create[i] = True

    l = max([len(file) for file in files])
    ncreate = sum(create)
    noverwrite = sum(overwrite)
    nskip = sum(skip)
    pcreate = not ((noverwrite > 0 and ncreate > 100) or ncreate > 300) or print_all
    pskip = (nskip <= 100) or print_all
    pcreate_message = ' (not printed)' if not pcreate else ''
    pskip_message = ' (not printed)' if not pskip else ''

    overview = []
    if ncreate > 0:
        overview += [String('create (->): {0:d}{1:s}'.format(ncreate, pcreate_message), color=theme['new']).format()]
    if noverwrite > 0:
        overview += [String('overwrite (=>): {0:d}'.format(noverwrite), color=theme['overwrite']).format()]
    if nskip > 0:
        overview += [String('skip (==): {0:d}{1:s}'.format(nskip, pskip_message), color=theme['skip']).format()]

    summary = []
    summary += ['- source : ' + os.path.normpath(src_dir)]
    summary += ['- dest   : ' + os.path.normpath(dest_dir)]
    summary += ['- ' + ', '.join(overview)]

    if ncreate + noverwrite <= 100 and print_summary:
        print('-----')
        print('\n'.join(summary))
        print('-----')

    if print_details:
        for i in range(n):
            if create[i] and pcreate:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=theme['bright']).format(),
                    String('->', color=theme['bright']).format(),
                    String(files[i], color=theme['new']).format()
                ))
            elif skip[i] and pskip:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=theme['skip']).format(),
                    String('==', color=theme['skip']).format(),
                    String(files[i], color=theme['skip']).format()
                ))
            elif overwrite[i]:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=theme['bright']).format(),
                    String('=>', color=theme['bright']).format(),
                    String(files[i], color=theme['overwrite']).format()
                ))

    if ncreate + noverwrite > 100 and print_summary:
        print('-----')
        print('\n'.join(summary))
        print('-----')

    if all(skip):
        return 0

    if not force:
        if not click.confirm('Proceed?'):
            raise IOError('Cancelled')

    MakeDirs(list(set([os.path.dirname(i) for i in dest])))

    i = np.argwhere(np.not_equal(skip, True)).ravel()
    src = np.array(src)[i]
    dest = np.array(dest)[i]
    files = np.array(files)[i]

    pbar = tqdm.trange(len(files), disable=quiet)
    for i in pbar:
        pbar.set_description(files[i])
        copy_function(src[i], dest[i])


def ShelephantCopySSH(
    copy_function,
    host,
    files,
    src_dir,
    dest_dir,
    checksum = False,
    quiet = False,
    force = False,
    print_details = True,
    print_summary = True,
    print_all = False,
    verbose = False,
    theme_name = 'none',
    yaml_hostinfo_src = None,
    yaml_hostinfo_dest = None,
    tempfilename = None):
    r'''
Get/send files.

:param function copy_function: Function to perform the copy. E.g. ``CopyFromRemote``.

:param str host: Host-name

:type files: list of str
:param files: Filenames (will be prepended by ``src_dir`` and ``dest_dir``).

:param str src_dir: The source directory.

:param str dest_dir: The destination directory.

:param bool checksum: Use checksum to skip files that are the same.

:param bool quiet: Proceed without printing progress.

:param bool force: Continue without prompt.

:param bool print_details: Print copy details.

:param bool print_summary: Print copy summary.

:param bool print_all: If ``False`` auto-truncation of long output is applied.

:type theme_name: str or None
:param theme_name: The name of the color-theme. See :py:mod:`shelephant.Theme`.

:type yaml_hostinfo_src: str, optional
:param yaml_hostinfo_src:
    Filename of hostinfo for the source, see :py:mod:`shelephant.cli.hostinfo`.
    Specify these files *only* to use precomputed checksums.

:type yaml_hostinfo_dest: str, optional
:param yaml_hostinfo_dest:
    Filename of hostinfo for the destination, see :py:mod:`shelephant.cli.hostinfo`.
    Specify these files *only* to use precomputed checksums.

:param str tempfilename: Filename for temporary file to use (e.g. for ``rsync``).
    '''

    assert type(files) == list

    if len(files) == 0:
        if not quiet:
            print('Nothing to do')
        return

    if copy_function in [CopyToRemote, RsyncToRemote]:
        to_remote = True
    elif copy_function in [CopyFromRemote, RsyncFromRemote]:
        to_remote = False
    else:
        raise IOError('Unknown copy function')

    if copy_function in [RsyncToRemote, RsyncFromRemote]:
        use_rsync = True
    else:
        use_rsync = False

    src = PrefixPaths(src_dir, files)
    dest = PrefixPaths(dest_dir, files)
    n = len(src)
    overwrite = [False for i in range(n)]
    create = [False for i in range(n)]
    skip = [False for i in range(n)]
    dest_exists = [False for i in range(n)]
    theme = Theme(theme_name.lower())

    if not os.path.isdir(dest_dir) and len(dest_dir) > 0:

        create = [True for i in range(n)]

    else:

        if checksum == True:
            src_checksums = GetChecksums(src, yaml_hostinfo_src, progress=not quiet)
            dest_checksums = GetChecksums(dest, yaml_hostinfo_dest, progress=not quiet)

        if to_remote:
            if yaml_hostinfo_dest:
                f = YamlRead(yaml_hostinfo_dest)['files']
                for i in range(n):
                    if files[i] in f:
                        dest_exists[i] = True
        elif not to_remote:
            for i in range(n):
                if os.path.isfile(dest[i]):
                    dest_exists[i] = True

        for i in range(n):
            if dest_exists[i]:
                if checksum:
                    if (src_checksums[i] == dest_checksums[i]) and (src_checksums[i] is not None):
                        skip[i] = True
                        continue
                overwrite[i] = True
                continue
            create[i] = True

    l = max([len(file) for file in files])
    ncreate = sum(create)
    noverwrite = sum(overwrite)
    nskip = sum(skip)
    pcreate = not ((noverwrite > 0 and ncreate > 100) or ncreate > 300) or print_all
    pskip = (nskip <= 100) or print_all
    pcreate_message = ' (not printed)' if not pcreate else ''
    pskip_message = ' (not printed)' if not pskip else ''

    overview = []
    if ncreate > 0:
        overview += [String('create (->): {0:d}{1:s}'.format(ncreate, pcreate_message), color=theme['new']).format()]
    if noverwrite > 0:
        overview += [String('overwrite (=>): {0:d}'.format(noverwrite), color=theme['overwrite']).format()]
    if nskip > 0:
        overview += [String('skip (==): {0:d}{1:s}'.format(nskip, pskip_message), color=theme['skip']).format()]

    summary = []
    if to_remote:
        summary += ['- to host        : ' + host]
        summary += ['- source (local) : ' + os.path.normpath(src_dir)]
        summary += ['- dest (remote)  : ' + os.path.normpath(dest_dir)]
    else:
        summary += ['- from host       : ' + host]
        summary += ['- source (remote) : ' + os.path.normpath(src_dir)]
        summary += ['- dest (local)    : ' + os.path.normpath(dest_dir)]
    summary += ['- ' + ', '.join(overview)]

    if ncreate + noverwrite <= 100 and print_summary:
        print('-----')
        print('\n'.join(summary))
        print('-----')

    if print_details:
        for i in range(n):
            if create[i] and pcreate:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=theme['bright']).format(),
                    String('->', color=theme['bright']).format(),
                    String(files[i], color=theme['new']).format()
                ))
            elif skip[i] and pskip:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=theme['skip']).format(),
                    String('==', color=theme['skip']).format(),
                    String(files[i], color=theme['skip']).format()
                ))
            elif overwrite[i]:
                print('{0:s} {1:s} {2:s}'.format(
                    String(files[i], width=l, color=theme['bright']).format(),
                    String('=>', color=theme['bright']).format(),
                    String(files[i], color=theme['overwrite']).format()
                ))

    if ncreate + noverwrite > 100 and print_summary:
        print('-----')
        print('\n'.join(summary))
        print('-----')

    if all(skip):
        return 0

    if not force:
        if not click.confirm('Proceed?'):
            raise IOError('Cancelled')

    if not to_remote:
        MakeDirs(list(set([os.path.dirname(i) for i in dest])))

    i = np.argwhere(np.not_equal(skip, True)).ravel()
    src = np.array(src)[i]
    dest = np.array(dest)[i]
    files = np.array(files)[i]

    if use_rsync:
        dest_dir = dest_dir if len(dest_dir) > 0 else '.'
        src_dir = src_dir if len(src_dir) > 0 else '.'
        return copy_function(host, src_dir, dest_dir, tempfilename, files, force, verbose, not quiet)

    pbar = tqdm.trange(len(files), disable=quiet)
    for i in pbar:
        pbar.set_description(files[i])
        copy_function(host, src[i], dest[i], verbose)


def Theme(theme=None):
    r'''
Return dictionary of colors.

.. code-block:: python

    {
        'new' : '...',
        'overwrite' : '...',
        'skip' : '...',
        'bright' : '...',
    }

:param str theme: Select color-theme.

:rtype: dict
    '''

    if theme == 'dark':
        return \
        {
            'new' : '1;32',
            'overwrite': '1;31',
            'skip' : '1;30',
            'bright' : '1;37',
        }

    return \
    {
        'new' : '',
        'overwrite': '',
        'skip' : '',
        'bright' : '',
    }


class String:
    r'''
Rich string.

.. note::

    All options are attributes, that can be modified at all times.

.. note::

    Available methods:

    *   ``A.format()`` :  Formatted string.
    *   ``str(A)`` : Unformatted string.
    *   ``A.isnumeric()`` : Return if the "data" is numeric.
    *   ``int(A)`` : Dummy integer.
    *   ``float(A)`` : Dummy float.

:type data: str, None
:param data: The data.

:type width: None, int
:param width: Print width (formatted print only).

:type color: None, str
:param color: Print color, e.g. "1;32" for bold green (formatted print only).

:type align: ``'<'``, ``'>'``
:param align: Print alignment (formatted print only).

:type dummy: 0, int, float
:param dummy: Dummy numerical value.

:methods:


    '''

    def __init__(self, data, width=None, align='<', color=None, dummy=0):

        self.data  = data
        self.width = width
        self.color = color
        self.align = align
        self.dummy = dummy

    def format(self):
        r'''
Return formatted string: align/width/color are applied.
        '''

        if self.width and self.color:
            fmt = '\x1b[{color:s}m{{0:{align:s}{width:d}.{width:d}s}}\x1b[0m'.format(**self.__dict__)
        elif self.width:
            fmt = '{{0:{align:s}{width:d}.{width:d}s}}'.format(**self.__dict__)
        elif self.color:
            fmt = '\x1b[{color:s}m{{0:{align:s}s}}\x1b[0m'.format(**self.__dict__)
        else:
            fmt = '{{0:{align:s}s}}'.format(**self.__dict__)

        return fmt.format(str(self))

    def isnumeric(self):
        r'''
Return if the "data" is numeric : always zero for this class.
        '''
        return False

    def __str__(self):
        return str(self.data)

    def __int__(self):
        return int(self.dummy)

    def __float__(self):
        return float(self.dummy)

    def __repr__(self):
        return str(self)

    def __lt__(self,other):
        return str(self) < str(other)




'''shelephant_get
    Copy files using earlier collected information on which files to copy from where.
    By default the back-end is ``rysnc -a --from-file="temp" host:source_dir dest_dir``.
    Alternatively ``scp -p host:source_file dest_file`` can be used.
    Typically, ``rsync`` will be faster, especially in copying a lot of small files.

:usage:

    shelephant_get [options]

    shelephant_get [options] <hostinfo.yaml>

:argument:

    YAML-file with host information. Default: shelephant_hostinfo.yaml

:options:

    -l, --local=N
        Add local 'host' information to use precomputed checksums.

    --scp
        Use ``scp`` instead of ``rysnc`` as backend.

    --temp=N
        Temporary filename to communicate with rsync. [default: shelephant_files.txt]

    --colors=M
        Select color scheme from: none, dark. [default: dark]

    -s, --summary
        Print summary (and no details unless specified).

    -d, --details
        Print details (and no summary unless specified).

    -f, --force
        Force overwrite of all existing (but not matching) files.

    --verbose
        Verbose all commands.

    -q, --quiet
        Do not print progress.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import os
import shutil

from .. import version
from .. import YamlRead
from .. import CopyFromRemote
from .. import RsyncFromRemote
from .. import ShelephantCopy
from .. import ShelephantCopySSH


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-l', '--local', required=False)
        parser.add_argument(      '--scp', required=False, action='store_true')
        parser.add_argument(      '--temp', required=False, default='shelephant_files.txt')
        parser.add_argument(      '--colors', required=False, default='dark')
        parser.add_argument('-s', '--summary', required=False, action='store_true')
        parser.add_argument('-d', '--details', required=False, action='store_true')
        parser.add_argument('-f', '--force', required=False, action='store_true')
        parser.add_argument(      '--verbose', required=False, action='store_true')
        parser.add_argument('-q', '--quiet', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('hostinfo', nargs='?', default='shelephant_hostinfo.yaml')
        args = parser.parse_args()

        source = args.hostinfo
        data = YamlRead(source)

        if 'host' not in data:

            ShelephantCopy(
                copy_function = shutil.copy2,
                files = data['files'],
                src_dir = data['prefix'],
                dest_dir = os.path.dirname(source),
                checksum = 'checksum' in data,
                quiet = args.quiet,
                force = args.force,
                print_details = not (args.force or args.summary) or args.details,
                print_summary = not (args.force or args.details) or args.summary,
                print_all = args.details,
                theme_name = args.colors.lower(),
                yaml_hostinfo_src = source,
                yaml_hostinfo_dest = args.local)

        elif args.scp:

            ShelephantCopySSH(
                copy_function = CopyFromRemote,
                host = data['host'],
                files = data['files'],
                src_dir = data['prefix'],
                dest_dir = os.path.dirname(source),
                checksum = 'checksum' in data,
                quiet = args.quiet,
                force = args.force,
                print_details = not (args.force or args.summary) or args.details,
                print_summary = not (args.force or args.details) or args.summary,
                print_all = args.details,
                verbose = args.verbose,
                theme_name = args.colors.lower(),
                yaml_hostinfo_src = source,
                yaml_hostinfo_dest = args.local)

        else:

            ShelephantCopySSH(
                copy_function = RsyncFromRemote,
                host = data['host'],
                files = data['files'],
                src_dir = data['prefix'],
                dest_dir = os.path.dirname(source),
                checksum = 'checksum' in data,
                quiet = args.quiet,
                force = args.force,
                print_details = not (args.force or args.summary) or args.details,
                print_summary = not (args.force or args.details) or args.summary,
                print_all = args.details,
                verbose = args.verbose,
                theme_name = args.colors.lower(),
                yaml_hostinfo_src = source,
                yaml_hostinfo_dest = args.local,
                tempfilename = args.temp)

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()



r'''shelephant_hostinfo
    Collect file information from location (on a remote host).

:usage:

    shelephant_hostinfo [options]

:options:

    -o, --output=N
        Output YAML-file. [default: shelephant_hostinfo.yaml].

    -r, --host=N
        Host-name.

    -p, --prefix=N
        Directory (on host) from which to copy.

    -f, --files=[N]
        YAML-file with list of files (on remote). [default: shelephant_dump.yaml]

    -c, --checksum=[N]
        YAML-file with checksums (on remote). [default: shelephant_checksum.yaml]

    --files-key=N
        Path in the YAML-file, separated by "/". [default: /]

    --checksum-key=N
        Path in the YAML-file, separated by "/". [default: /]

    --ignore
        Skip basic checks.

    -i, --input=[N]
        Use host and prefix from existing file. [default: shelephant_hostinfo.yaml]

    --remove=N...
        Remove paths (and checksums) from existing file.

    --force
        Overwrite output file without prompt.

    --verbose
        Verbose all commands.

    -h, --help
        Show help.

    --version
        Show version.


(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import os
import tempfile
import shutil
import numpy as np

from .. import version
from .. import YamlGetItem
from .. import YamlDump
from .. import YamlRead
from .. import CopyFromRemote


def remove(data, rm):

    files = np.array(data['files'])
    sorter = np.argsort(files)
    files = files[sorter]
    rm = np.sort(rm)

    test = np.in1d(rm, files)
    if not np.all(np.in1d(rm, files)):
        raise IOError('One or more remove paths not found')

    idx = np.searchsorted(files, rm)
    keep = np.setdiff1d(np.arange(len(files)), sorter[idx])

    files = data['files']
    checksum = data['checksum']
    data['files'] = [files[i] for i in keep]
    data['checksum'] = [checksum[i] for i in keep]

    return data


def main():

    try:

        # Parse command-line arguments

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-o', '--output', required=False, default='shelephant_hostinfo.yaml')
        parser.add_argument('-r', '--host', required=False, default=None)
        parser.add_argument('-p', '--prefix', required=False, default=None)
        parser.add_argument('-f', '--files', required=False, default=None, nargs='?', const='shelephant_dump.yaml')
        parser.add_argument('-c', '--checksum', required=False, default=None, nargs='?', const='shelephant_checksum.yaml')
        parser.add_argument(      '--files-key', required=False, default='/')
        parser.add_argument(      '--checksum-key', required=False, default='/')
        parser.add_argument(      '--ignore', required=False, action='store_true')
        parser.add_argument('-i', '--input', required=False, default=None, nargs='?', const='shelephant_hostinfo.yaml')
        parser.add_argument(      '--remove', required=False, nargs='*')
        parser.add_argument(      '--force', required=False, action='store_true')
        parser.add_argument(      '--verbose', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        args = parser.parse_args()

        # Separate mode: remove paths and quit

        if args.remove:
            filename = args.output
            if args.input:
                filename = args.input
            data = YamlRead(filename)
            data = remove(data, args.remove)
            YamlDump(args.output, data, args.force)
            return 0

        # Extract basic information from command-line arguments

        data = {}

        if args.host:
            data['host'] = args.host

        if args.prefix:
            data['prefix'] = args.prefix

        if args.input:
            overwrite = YamlRead(args.input)
            for item in ['host', 'prefix']:
                if item in overwrite:
                    data[item] = overwrite[item]

        # Basic IO-checks

        if not args.ignore:
            if 'host' in data and 'prefix' not in data:
                raise IOError('Specify hostname and prefix')

        if 'host' not in data and 'prefix' not in data:
            for item in [args.files, args.checksum]:
                if item:
                    data['prefix'] = os.path.dirname(item)
                    break

        # Create temporary file to download to

        if 'host' in data:
            temp_dir = tempfile.mkdtemp()
            temp_file = os.path.join(temp_dir, 'shelephant_hostinfo.yaml')

        # Read files and checksums

        args_dict = vars(args)

        for item in ['files', 'checksum']:

            if args_dict[item]:

                key = list(filter(None, args_dict[item + '_key'].split('/')))
                filename = args_dict[item]

                if 'host' in data:
                    CopyFromRemote(
                        data['host'],
                        os.path.join(data['prefix'], filename),
                        temp_file,
                        args.verbose)
                    filename = temp_file

                data[item] = YamlGetItem(filename, key)

        # Run basic checks

        if not args.ignore:
            for item in ['prefix']:
                if item not in data:
                    raise IOError('Please specify {0:s}'.format(item))
            if 'checksum' in data:
                if len(data['files']) != len(data['checksum']):
                    raise IOError('Number of checksums does not match number of files')

        # Clean-up: remove temporary file

        if 'host' in data:
            shutil.rmtree(temp_dir)

        # Write output

        if 'files' not in data:
            data['files'] = []

        YamlDump(args.output, data, args.force)

    except Exception as e:

        print(e)
        return 1

if __name__ == '__main__':

    main()



'''shelephant_send
    Copy files to a remote, using earlier collected information on which files to copy where.
    By default the back-end is ``rysnc -a --from-file="temp" source_dir host:dest_dir``.
    Alternatively ``scp -p source_file host:dest_file`` can be used.
    Typically, ``rsync`` will be faster, especially in copying a lot of small files.

:usage:

    shelephant_send [options]

    shelephant_send [options] <files.yaml> <hostinfo.yaml>

:arguments:

    <files.yaml>
        YAML-file with files to send. Default: shelephant_dump.yaml

    <hostinfo.yaml>
        YAML-file with host information. Default: shelephant_hostinfo.yaml

:options:

    -k, --key=N
        Path in <files.yaml>, separated by "/". [default: /]

    -l, --local=N
        Add local 'host' information to use precomputed checksums.

    --scp
        Use ``scp`` instead of ``rysnc`` as backend.

    --temp=N
        Temporary filename to communicate with rsync. [default: shelephant_files.txt]

    --colors=M
        Select color scheme from: none, dark. [default: dark]

    -s, --summary
        Print summary (and no details unless specified).

    -d, --details
        Print details (and no summary unless specified).

    -f, --force
        Force overwrite of all existing (but not matching) files.

    --verbose
        Verbose all commands.

    -q, --quiet
        Do not print progress.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import os
import shutil

from .. import version
from .. import YamlRead
from .. import YamlGetItem
from .. import CopyToRemote
from .. import RsyncToRemote
from .. import ShelephantCopy
from .. import ShelephantCopySSH


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-k', '--key', required=False, default='/')
        parser.add_argument('-l', '--local', required=False)
        parser.add_argument(      '--scp', required=False, action='store_true')
        parser.add_argument(      '--temp', required=False, default='shelephant_files.txt')
        parser.add_argument(      '--colors', required=False, default='dark')
        parser.add_argument('-s', '--summary', required=False, action='store_true')
        parser.add_argument('-d', '--details', required=False, action='store_true')
        parser.add_argument('-f', '--force', required=False, action='store_true')
        parser.add_argument(      '--verbose', required=False, action='store_true')
        parser.add_argument('-q', '--quiet', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('args', nargs=2, default=['shelephant_dump.yaml', 'shelephant_hostinfo.yaml'])
        args = parser.parse_args()

        source = args.args[0]
        hostinfo = args.args[1]
        data = YamlRead(hostinfo)
        key = list(filter(None, args.key.split('/')))
        files = YamlGetItem(source, key)
        src_dir = os.path.dirname(source)
        dest_dir = data['prefix']

        if 'host' not in data:

            ShelephantCopy(
                copy_function = shutil.copy2,
                files = files,
                src_dir = src_dir,
                dest_dir = data['prefix'],
                checksum = 'checksum' in data,
                quiet = args.quiet,
                force = args.force,
                print_details = not (args.force or args.summary) or args.details,
                print_summary = not (args.force or args.details) or args.summary,
                print_all = args.details,
                theme_name = args.colors.lower(),
                yaml_hostinfo_src = args.local,
                yaml_hostinfo_dest = hostinfo)

        elif args.scp:

            ShelephantCopySSH(
                copy_function = CopyToRemote,
                host = data['host'],
                files = files,
                src_dir = src_dir,
                dest_dir = dest_dir,
                checksum = 'checksum' in data,
                quiet = args.quiet,
                force = args.force,
                print_details = not (args.force or args.summary) or args.details,
                print_summary = not (args.force or args.details) or args.summary,
                print_all = args.details,
                verbose = args.verbose,
                theme_name = args.colors.lower(),
                yaml_hostinfo_src = args.local,
                yaml_hostinfo_dest = hostinfo)

        else:

            ShelephantCopySSH(
                copy_function = RsyncToRemote,
                host = data['host'],
                files = files,
                src_dir = src_dir,
                dest_dir = dest_dir,
                checksum = 'checksum' in data,
                quiet = args.quiet,
                force = args.force,
                print_details = not (args.force or args.summary) or args.details,
                print_summary = not (args.force or args.details) or args.summary,
                print_all = args.details,
                verbose = args.verbose,
                theme_name = args.colors.lower(),
                yaml_hostinfo_src = args.local,
                yaml_hostinfo_dest = hostinfo,
                tempfilename = args.temp)

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()



import unittest
import subprocess
import os
import shutil
import numpy as np

from shelephant import YamlGetItem
from shelephant import YamlRead
from shelephant import YamlDump
from shelephant import GetDeepestPaths


def run(cmd, verbose=False):
    return subprocess.check_output(cmd, shell=True).decode('utf-8')


class Test_tools(unittest.TestCase):

    def test_FlattenList(self):

        from shelephant import FlattenList

        arg = [1, [2, 2, 2], 4]
        ret = [1, 2, 2, 2, 4]

        self.assertEqual(ret, FlattenList(arg))

    def test_Squash(self):

        from shelephant import Squash

        arg = {'foo': [1, 2], 'bar': {'foo': [3, 4], 'bar': 5}}
        ret = [1, 2, 3, 4, 5]

        self.assertEqual(ret, Squash(arg))

    def test_deepest_dirs(self):

        dirnames = [
            "foo/bar",
            "foo/bar/mydir",
            "bar/mydir",
            "bar" ,
            "bar/mydir2",
            "foo/bar/mydir/deep",
            "foo/bar/mydir/deep/deeper",
            "foo/bar/mydir/deep/also",
            "foo/shallow"]

        ret = GetDeepestPaths(dirnames)

        d = [
            "bar/mydir",
            "bar/mydir2",
            "foo/bar/mydir/deep/deeper",
            "foo/bar/mydir/deep/also",
            "foo/shallow"]

        self.assertEqual(sorted(ret), sorted(d))


class Test_checksum(unittest.TestCase):

    def test_basic(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_checksum -f -q shelephant_dump.yaml')
        data = YamlGetItem('shelephant_checksum.yaml')

        keys = [
            '2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae',
            'fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9',
        ]

        self.assertEqual(data, keys)

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('shelephant_dump.yaml')
        os.remove('shelephant_checksum.yaml')

    def test_hybrid(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f foo.txt')
        output = run('shelephant_checksum -f -q')
        output = run('shelephant_hostinfo --force -f -c')

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_checksum -f -q -l shelephant_hostinfo.yaml')
        data = YamlGetItem('shelephant_checksum.yaml')

        keys = [
            '2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae',
            'fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9',
        ]

        self.assertEqual(data, keys)

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('shelephant_dump.yaml')
        os.remove('shelephant_checksum.yaml')
        os.remove('shelephant_hostinfo.yaml')

    def test_recursive(self):

        letters = ['a', 'b', 'c', 'd', 'e', 'f', 'g']

        for letter in letters:
            with open('{0:s}.txt'.format(letter), 'w') as file:
                file.write(letter)

        files = ['{0:s}.txt'.format(letter) for letter in letters]

        keys = [
            'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb', # a
            '3e23e8160039594a33894f6564e1b1348bbd7a0088d42c4acb73eeaed59c009d', # b
            '2e7d2c03a9507ae265ecf5b5356885a53393a2029d241394997265a1a25aefc6', # c
            '18ac3e7343f016890c510e93f935261169d9e3f565436429830faf0934f4f8e4', # d
            '3f79bb7b435b05321651daefd374cdc681dc06faa65e374e38337b88ca046dea', # e
            '252f10c83610ebca1a059c0bae8255eba2f95be4d1d7bcfa89d7248a82d9f111', # f
            'cd0aa9856147b6c5b4ff2b7dfee5da20aa38253099ef1b4a64aced233c9afe29', # g
        ]

        output = run('shelephant_dump -f -s {0:s}'.format(' '.join(files)))
        output = run('shelephant_checksum -f -q')

        data = YamlGetItem('shelephant_checksum.yaml')

        self.assertEqual(data, keys)

        ifiles = np.arange(len(files))
        np.random.shuffle(ifiles)

        output = run('shelephant_dump -f {0:s}'.format(' '.join([files[i] for i in ifiles[:3]])))
        output = run('shelephant_checksum -f -q')
        output = run('shelephant_hostinfo --force -f -c')

        np.random.shuffle(ifiles)

        output = run('shelephant_dump -f {0:s}'.format(' '.join([files[i] for i in ifiles])))
        output = run('shelephant_checksum -f -q -l shelephant_hostinfo.yaml')
        output = run('shelephant_hostinfo --force -f -c')

        output = run('shelephant_dump -f -s {0:s}'.format(' '.join(files)))
        output = run('shelephant_checksum -f -q -l shelephant_hostinfo.yaml')

        data = YamlGetItem('shelephant_checksum.yaml')

        self.assertEqual(data, keys)

        for file in files:
            os.remove(file)

        os.remove('shelephant_dump.yaml')
        os.remove('shelephant_checksum.yaml')
        os.remove('shelephant_hostinfo.yaml')


class Test_dump(unittest.TestCase):

    def test_basic(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        for dirname in ['mydir']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mydir')

        with open('mydir/foo.txt', 'w') as file:
            file.write('foo')

        with open('mydir/bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f -s -o dump_1.yaml foo.txt bar.txt')
        output = run('shelephant_dump -f -s -o dump_2.yaml *.txt')
        output = run('shelephant_dump -f -s -o mydir/dump_3.yaml mydir/*.txt')

        with open('dump_1.yaml', 'r') as file:
            dump_1 = file.read()

        with open('dump_2.yaml', 'r') as file:
            dump_2 = file.read()

        with open('mydir/dump_3.yaml', 'r') as file:
            dump_3 = file.read()

        self.assertEqual(dump_1, dump_2)
        self.assertEqual(dump_1, dump_3)

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('dump_1.yaml')
        os.remove('dump_2.yaml')
        shutil.rmtree('mydir')

    def test_append(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        with open('foo.pdf', 'w') as file:
            file.write('foo')

        with open('bar.pdf', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_dump -a foo.pdf bar.pdf')

        self.assertEqual(YamlRead('shelephant_dump.yaml'), ['foo.txt', 'bar.txt', 'foo.pdf', 'bar.pdf'])

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('foo.pdf')
        os.remove('bar.pdf')
        os.remove('shelephant_dump.yaml')


class Test_extract(unittest.TestCase):

    def test_single_path(self):

        data = {
            'foo' : ['foo.txt', 'bar.txt'],
            'bar' : ['foo.pdf', 'bar.pdf'],
            'key' : ['foo.key', 'bar.key'],
        }

        YamlDump('dump.yaml', data, force=True)

        output = run('shelephant_extract -f dump.yaml "foo"')

        self.assertEqual(YamlRead('dump.yaml'), ['foo.txt', 'bar.txt'])

        os.remove('dump.yaml')

    def test_multiple_paths(self):

        data = {
            'foo' : ['foo.txt', 'bar.txt'],
            'bar' : ['foo.pdf', 'bar.pdf'],
            'key' : ['foo.key', 'bar.key'],
            'sub' : {
                'foo' : ['foo.txt', 'bar.txt'],
                'bar' : ['foo.pdf', 'bar.pdf'],
                'key' : ['foo.key', 'bar.key'],
            },
        }

        YamlDump('dump.yaml', data, force=True)

        output = run('shelephant_extract -f dump.yaml "/sub/foo" "foo"')

        self.assertEqual(YamlRead('dump.yaml'), {'foo': ['foo.txt', 'bar.txt'], 'sub': {'foo': ['foo.txt', 'bar.txt']}})

        os.remove('dump.yaml')

    def test_multiple_paths_squash(self):

        data = {
            'foo' : ['foo.txt', 'bar.txt'],
            'bar' : ['foo.pdf', 'bar.pdf'],
            'key' : ['foo.key', 'bar.key'],
            'sub' : {
                'foo' : ['foo2.txt', 'bar2.txt'],
                'bar' : ['foo2.pdf', 'bar2.pdf'],
                'key' : ['foo2.key', 'bar2.key'],
            },
        }

        YamlDump('dump.yaml', data, force=True)

        output = run('shelephant_extract -f --squash dump.yaml "/sub/foo" "foo"')

        self.assertEqual(YamlRead('dump.yaml'), ['foo2.txt', 'bar2.txt', 'foo.txt', 'bar.txt'])

        os.remove('dump.yaml')


class Test_merge(unittest.TestCase):

    def test_basic(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -o main.yaml foo.txt')
        output = run('shelephant_dump -o branch.yaml bar.txt')
        output = run('shelephant_merge -f branch.yaml main.yaml')

        self.assertEqual(YamlRead('main.yaml'), ['foo.txt', 'bar.txt'])

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('main.yaml')
        os.remove('branch.yaml')

        os.mkdir('dira')
        os.mkdir('dirb')

        with open('dira/foo.txt', 'w') as file:
            file.write('foo')

        with open('dira/bar.txt', 'w') as file:
            file.write('bar')

        with open('dirb/foo.txt', 'w') as file:
            file.write('foo')

        with open('dirb/bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -o dira/dump.yaml dira/foo.txt dira/bar.txt')
        output = run('shelephant_dump -o dirb/dump.yaml dirb/foo.txt dirb/bar.txt')
        output = run('shelephant_merge -f dira/dump.yaml dirb/dump.yaml')

        self.assertEqual(YamlRead('dirb/dump.yaml'), ['foo.txt', 'bar.txt', '../dira/foo.txt', '../dira/bar.txt'])

        shutil.rmtree('dira')
        shutil.rmtree('dirb')


class Test_hostinfo(unittest.TestCase):

    def test_basic(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f -s -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml')
        output = run('shelephant_get -f -q mydest/hostinfo.yaml')
        output = run('shelephant_dump -f -s -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(YamlRead('mysrc/files.yaml'), YamlRead('mydest/files.yaml'))
        self.assertEqual(YamlRead('mysrc/checksum.yaml'), YamlRead('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')


    def test_remove(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        keys = [
            '2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae',
            'fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9',
        ]

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_checksum -f -q')
        output = run('shelephant_hostinfo --force -f -c')
        output = run('shelephant_hostinfo --force --remove bar.txt')
        data = YamlRead('shelephant_hostinfo.yaml')

        self.assertEqual(data['files'], ['foo.txt'])
        self.assertEqual(data['checksum'], [keys[0]])

        os.remove('foo.txt')
        os.remove('bar.txt')
        os.remove('shelephant_dump.yaml')
        os.remove('shelephant_checksum.yaml')
        os.remove('shelephant_hostinfo.yaml')


class Test_get(unittest.TestCase):

    def test_basic(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt -> foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml')
        output = run('shelephant_get -f -d -q --colors none mydest/hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(YamlRead('mysrc/files.yaml'), YamlRead('mydest/files.yaml'))
        self.assertEqual(YamlRead('mysrc/checksum.yaml'), YamlRead('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')

    def test_partial(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        shutil.copy('mysrc/foo.txt', 'mydest/foo.txt')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt == foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml')
        output = run('shelephant_get -f -d -q --colors none mydest/hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(YamlRead('mysrc/files.yaml'), YamlRead('mydest/files.yaml'))
        self.assertEqual(YamlRead('mysrc/checksum.yaml'), YamlRead('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')

    def test_partial_localchecksum(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        with open('mysrc/car.txt', 'w') as file:
            file.write('car')

        with open('mysrc/dog.txt', 'w') as file:
            file.write('dog')

        shutil.copy('mysrc/foo.txt', 'mydest/foo.txt')
        shutil.copy('mysrc/dog.txt', 'mydest/dog.txt')

        operations = [
            'bar.txt -> bar.txt',
            'car.txt -> car.txt',
            'dog.txt == dog.txt',
            'foo.txt == foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_hostinfo -o mydest/hostinfo.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml')
        output = run('shelephant_hostinfo -o mydest/local.yaml -f mydest/files.yaml -c mydest/checksum.yaml')
        output = run('shelephant_get -f -d -q --colors none -l mydest/local.yaml mydest/hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump -f -s -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(YamlRead('mysrc/files.yaml'), YamlRead('mydest/files.yaml'))
        self.assertEqual(YamlRead('mysrc/checksum.yaml'), YamlRead('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')


class Test_send(unittest.TestCase):

    def test_basic(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        with open('mydest/foobar.txt', 'w') as file:
            file.write('foobar')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt -> foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_hostinfo --force -o hostinfo.yaml -f mydest/files.yaml -c mydest/checksum.yaml')
        output = run('shelephant_send -f -d -q --colors none mysrc/files.yaml hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        os.remove('mydest/foobar.txt')

        output = run('shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(YamlRead('mysrc/files.yaml'), YamlRead('mydest/files.yaml'))
        self.assertEqual(YamlRead('mysrc/checksum.yaml'), YamlRead('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('hostinfo.yaml')

    def test_empty_remote(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt -> foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_hostinfo --force -o hostinfo.yaml -p mydest')
        output = run('shelephant_send -f -d -q --colors none mysrc/files.yaml hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_checksum -f -q -o mysrc/checksum.yaml mysrc/files.yaml')

        self.assertEqual(YamlRead('mysrc/files.yaml'), YamlRead('mydest/files.yaml'))
        self.assertEqual(YamlRead('mysrc/checksum.yaml'), YamlRead('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('hostinfo.yaml')


    def test_partial(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        shutil.copy('mysrc/foo.txt', 'mydest/foo.txt')

        operations = [
            'bar.txt -> bar.txt',
            'foo.txt == foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_hostinfo --force -o hostinfo.yaml -f mydest/files.yaml -c mydest/checksum.yaml')
        output = run('shelephant_send -f -d -q --colors none mysrc/files.yaml hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(YamlRead('mysrc/files.yaml'), YamlRead('mydest/files.yaml'))
        self.assertEqual(YamlRead('mysrc/checksum.yaml'), YamlRead('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('hostinfo.yaml')

    def test_partial_localchecksum(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        with open('mysrc/car.txt', 'w') as file:
            file.write('car')

        with open('mysrc/dog.txt', 'w') as file:
            file.write('dog')

        shutil.copy('mysrc/foo.txt', 'mydest/foo.txt')
        shutil.copy('mysrc/dog.txt', 'mydest/dog.txt')

        operations = [
            'bar.txt -> bar.txt',
            'car.txt -> car.txt',
            'dog.txt == dog.txt',
            'foo.txt == foo.txt',
        ]

        output = run('shelephant_dump --sort -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')
        output = run('shelephant_hostinfo --force -o hostinfo.yaml -f mydest/files.yaml -c mydest/checksum.yaml')
        output = run('shelephant_hostinfo --force -o local.yaml -f mysrc/files.yaml -c mysrc/checksum.yaml')
        output = run('shelephant_send -f -d -q --colors none -l local.yaml mysrc/files.yaml hostinfo.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), operations)

        output = run('shelephant_dump -f --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -f -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(YamlRead('mysrc/files.yaml'), YamlRead('mydest/files.yaml'))
        self.assertEqual(YamlRead('mysrc/checksum.yaml'), YamlRead('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')
        os.remove('hostinfo.yaml')
        os.remove('local.yaml')


class Test_mv(unittest.TestCase):

    def test_basic(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_mv -f -q mysrc/files.yaml mydest')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(YamlRead('mysrc/files.yaml'), YamlRead('mydest/files.yaml'))
        self.assertEqual(YamlRead('mysrc/checksum.yaml'), YamlRead('mydest/checksum.yaml'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')


class Test_cp(unittest.TestCase):

    def test_basic(self):

        for dirname in ['mysrc', 'mydest']:
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

        os.mkdir('mysrc')
        os.mkdir('mydest')

        with open('mysrc/foo.txt', 'w') as file:
            file.write('foo')

        with open('mysrc/bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -o mysrc/files.yaml mysrc/*.txt')
        output = run('shelephant_checksum -q -o mysrc/checksum.yaml mysrc/files.yaml')
        output = run('shelephant_cp -f -q mysrc/files.yaml mydest')
        output = run('shelephant_dump --sort -o mydest/files.yaml mydest/*.txt')
        output = run('shelephant_checksum -q -o mydest/checksum.yaml mydest/files.yaml')

        self.assertEqual(YamlRead('mysrc/files.yaml'), YamlRead('mydest/files.yaml'))
        self.assertEqual(YamlRead('mysrc/checksum.yaml'), YamlRead('mydest/checksum.yaml'))
        self.assertTrue(os.path.isfile('mysrc/foo.txt'))
        self.assertTrue(os.path.isfile('mysrc/bar.txt'))

        shutil.rmtree('mysrc')
        shutil.rmtree('mydest')


class Test_rm(unittest.TestCase):

    def test_basic(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_rm -f shelephant_dump.yaml')

        self.assertFalse(os.path.isfile('foo.txt'))
        self.assertFalse(os.path.isfile('bar.txt'))

        os.remove('shelephant_dump.yaml')


class Test_parse(unittest.TestCase):

    def test_basic(self):

        with open('foo.txt', 'w') as file:
            file.write('foo')

        with open('bar.txt', 'w') as file:
            file.write('bar')

        output = run('shelephant_dump -f foo.txt bar.txt')
        output = run('shelephant_parse shelephant_dump.yaml')

        self.assertEqual(list(filter(None, output.split('\n'))), ['- foo.txt', '- bar.txt'])

        os.remove('shelephant_dump.yaml')
        os.remove('foo.txt')
        os.remove('bar.txt')


if __name__ == '__main__':

    unittest.main()



'''shelephant_extract
    Extract a field from a YAML-file.

    Unless you use --no-path, the function assumes that all data are paths,
    and changes all relative paths from being relative to <input.yaml>
    to being relative to --output.

:usage:

    shelephant_extract [options] <input.yaml> [<key>...]

:arguments:

    <input.yaml>
        The file to read.

    <key>
        The keys to read from the file.

:options:

    -o, --output=N
        Output file. (default: <input.yaml>)

    --no-path
        Do not interpret data as paths.

    -s, --squash
        Squash fields into a single field.

    -f, --force
        Overwrite output file without prompt.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
'''

import argparse
import os
import mergedeep
import functools

from .. import version
from .. import YamlGetItem
from .. import YamlDump
from .. import ChangeRootOfRelativePaths
from .. import Squash


def main():

    try:

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument('-o', '--output', required=False)
        parser.add_argument(      '--no-path', required=False, action='store_true')
        parser.add_argument('-s', '--squash', required=False, action='store_true')
        parser.add_argument('-f', '--force', required=False, action='store_true')
        parser.add_argument('-v', '--version', action='version', version=version)
        parser.add_argument('input')
        parser.add_argument('key', nargs='*')
        args = parser.parse_args()

        input_dir = os.path.dirname(args.input)
        output = args.output if args.output else args.input
        output_dir = os.path.dirname(output)
        data = {}

        if len(args.key) == 0:
            args.key = ['/']

        for key in args.key:
            key = list(filter(None, key.split('/')))
            files = YamlGetItem(args.input, key)
            if not args.no_path:
                files = ChangeRootOfRelativePaths(files, input_dir, output_dir)
            if len(args.key) == 1:
                YamlDump(output, files, args.force)
                return 0
            container = functools.reduce(lambda x, y: {y: x}, key[:-1], {key[-1]: files})
            mergedeep.merge(data, container)

        if args.squash:
            data = Squash(data)

        YamlDump(output, data, args.force)

    except Exception as e:

        print(e)
        return 1


if __name__ == '__main__':

    main()
